#include <stdio.h>
#include <stdlib.h>

#include <sys/time.h>
#include <stdlib.h>
#include <pthread.h>
#include <unistd.h>
#include <assert.h>
#include <sys/mman.h>

#include "mthread.h"
#include "atomic.h"

int numThreads;
int tid_gen = 0;
pthread_mutex_t global_mutex = PTHREAD_MUTEX_INITIALIZER;


int __thread tid;

void *mthread_start_routine(void *void_args)
{
    void *ret;
    mthread_arg_t *args = (mthread_arg_t *) void_args;

    pthread_mutex_lock(&global_mutex);
    tid = tid_gen++;
    pthread_mutex_unlock(&global_mutex);

    ret = args->start_routine(args->args);
    free(args);

    pthread_mutex_lock(&global_mutex);
    numThreads--;

    if (tid == numThreads)
        tid_gen--;

    if (numThreads == 1)
        tid_gen = 1;

    pthread_mutex_unlock(&global_mutex);

    return ret;
}

int mthread_create(pthread_t *thread, const pthread_attr_t *attr,
        void *(*start_routine)(void*), void *args)
{
    mthread_arg_t *newargs = (mthread_arg_t *) malloc(sizeof(mthread_arg_t));
    newargs->start_routine = start_routine;
    newargs->args = args;

    pthread_mutex_lock(&global_mutex);
    if (numThreads == 0)
    {
        tid = tid_gen++;
        numThreads++;
    }
    numThreads++;
    pthread_mutex_unlock(&global_mutex);

    assert(numThreads < MAXTHREADS);

    return pthread_create(thread, attr, mthread_start_routine, (void *) newargs);
}


int mthread_mutex_init(mthread_mutex_t *mutex, const void *ignore)
{
#if 1
    mutex->nodes = (node_t*) ((void *) mutex->node_data + 
            (L2_CACHELINE_SIZE - (((int) mutex->node_data) % L2_CACHELINE_SIZE)));
    assert((((int) mutex->nodes) % L2_CACHELINE_SIZE) == 0);
#else
    mutex->nodes = mutex->node_data; 
#endif


    mutex->tail = NULL;
    return 0;
}


int mthread_mutex_lock(mthread_mutex_t *mutex) 
{
#ifdef SMT
    node_t *node = &mutex->nodes[(tid/2)*4 + (tid % 2)];
#else
    node_t *node = &mutex->nodes[tid];
#endif
    node_t *pred;
    
    node->next = NULL;
    pred = (node_t *) fetch_and_store((int *) &mutex->tail, (int) node);

    if (pred != NULL)
    {
        node->status = 1;
        pred->next = node;

        spin(&node->status);
    }

    return 0;
}

int mthread_mutex_unlock(mthread_mutex_t *mutex) 
{
#ifdef SMT
    node_t *node = &mutex->nodes[(tid/2)*4 + (tid % 2)];
#else
    node_t *node = &mutex->nodes[tid];
#endif

    asm __volatile__ ("" : "=m" (node->next) : "m" (node->next));
    
    // If there is no one waiting, we have to atomically write to the queue tail
    // pointer.
    if (node->next == NULL)
    {
        if (compare_and_swap((int *) &mutex->tail, (int) node, 0))
            return;

        // Wait for successor to make themselves known (should be quick)
        while (node->next == NULL)
        {
            asm __volatile__ ("" : "=m" (node->next) : "m" (node->next));
        }
    }

    node->next->status = 0;
}
