#ifndef MTHREAD_H
#define MTHREAD_H

#include <pthread.h>

#define MAXTHREADS 128

#define L1_CACHELINE_SIZE 32
#define L2_CACHELINE_SIZE 32

typedef struct
{
    void *(*start_routine)(void*);
    void *args;
} mthread_arg_t;

typedef struct node_
{
    int status;
    struct node_ *next;
#ifdef SMT
    char padding[L1_CACHELINE_SIZE - sizeof(int) - sizeof(struct node_ *)];
#else
    char padding[L2_CACHELINE_SIZE - sizeof(int) - sizeof(struct node_ *)];
#endif
} node_t;

typedef struct
{
    node_t node_data[MAXTHREADS + 2];
    node_t *nodes;
    node_t *tail;
} mthread_mutex_t;

int mthread_create(pthread_t *thread, const pthread_attr_t *attr, void *(*start_routine)(void*), void *args);
int mthread_mutex_init(mthread_mutex_t *mutex, const void *ignore);
int mthread_mutex_lock(mthread_mutex_t *mutex); 
int mthread_mutex_unlock(mthread_mutex_t *mutex); 

#endif // !def MTHREAD_H
